package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kbox2d.dynamics.Body
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import kotlin.math.sqrt

/// Friction mixing law. The idea is to allow either fixture to drive the friction to zero.
/// For example, anything slides on ice.
fun b2MixFriction(friction1: Double, friction2: Double): Double {
	return sqrt(friction1 * friction2)
}

/// Restitution mixing law. The idea is allow for anything to bounce off an inelastic surface.
/// For example, a superball bounces on anything.
fun b2MixRestitution(restitution1: Double, restitution2: Double): Double {
	return if (restitution1 > restitution2) restitution1 else restitution2
}

/// Restitution mixing law. This picks the lowest value.
fun b2MixRestitutionThreshold(threshold1: Double, threshold2: Double): Double {
	return if (threshold1 < threshold2) threshold1 else threshold2
}

data class ContactRegister(
	val primary: Boolean,
)

/**
 * A contact edge is used to connect bodies and contacts together
 * in a contact graph where each body is a node and each contact
 * is an edge. A contact edge belongs to a doubly linked list
 * maintained in each attached body. Each contact has two contact
 * nodes, one for each attached body.
 */
data class ContactEdge(
	val other: Body, ///< provides quick access to the other body attached.
	val contact: AbstractContact, ///< the contact
	var prev: ContactEdge? = null, ///< the previous contact edge in the body's contact list
	var next: ContactEdge? = null, ///< the next contact edge in the body's contact list
)

enum class ContactFlags(val bitmask: Int) {
	ISLAND(0x1),
	TOUCHING(0x2),
	ENABLED(0x4),
	FILTER(0x8),
	BULLET_HIT(0x10),
	TOI(0x20);

	fun isit(other: Int): Boolean {
		return (other and bitmask) == bitmask
	}

	fun update(other: Int, state: Boolean): Int {
		if (state)
			return or(other)
		else
			return not(other)
	}

	fun and(other: Int): Int {
		return other and bitmask
	}

	fun or(other: Int): Int {
		return other or bitmask
	}

	fun not(other: Int): Int {
		return other and bitmask.inv()
	}
}
