package ru.dbotthepony.kbox2d.collision

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

typealias b2Pair = Pair<Int, Int>

const val e_nullProxy = -1

private class IntArrayList {
	private var buffer = IntArray(16)
	var size: Int = 0
		private set

	operator fun set(index: Int, value: Int) {
		require(index in 0 until size) { "Index out of bounds: $index" }
		buffer[index] = value
	}

	operator fun get(index: Int): Int {
		require(index in 0 until size) { "Index out of bounds: $index" }
		return buffer[index]
	}

	fun add(value: Int) {
		if (size == buffer.size) {
			val old = buffer
			buffer = IntArray(size * 2)

			for (i in 0 until size) {
				buffer[i] = old[i]
			}
		}

		buffer[size++] = value
	}
}

/**
 * The broad-phase is used for computing pairs and performing volume queries and ray casts.
 * This broad-phase does not persist pairs. Instead, this reports potentially new pairs.
 * It is up to the client to consume the new pairs and to track subsequent overlap.
 */
class BroadPhase : IProxieable, IMovable {
	private val moveBuffer = IntArrayList()
	private val pairBuffer = ArrayList<b2Pair>()
	private var moveCount = 0

	private val tree = DynamicTree()

	/**
	 * Get the number of proxies.
	 */
	var proxyCount: Int = 0
		private set

	/**
	 * Call to trigger a re-processing of it's pairs on the next call to UpdatePairs.
	 */
	fun touchProxy(proxyID: Int) {
		bufferMove(proxyID)
	}

	override fun createProxy(aabb: AABB, userData: Any?): Int {
		val proxyId = tree.createProxy(aabb, userData)
		proxyCount++
		bufferMove(proxyId)
		return proxyId
	}

	override fun destroyProxy(proxyID: Int) {
		unBufferMove(proxyID)
		proxyCount--
		tree.destroyProxy(proxyID)
	}

	override fun moveProxy(proxyID: Int, aabb: AABB, displacement: Vector2d): Boolean {
		if (tree.moveProxy(proxyID, aabb, displacement)) {
			bufferMove(proxyID)
			return true
		}

		return false
	}

	private fun bufferMove(proxyId: Int) {
		if (moveBuffer.size > moveCount) {
			moveBuffer[moveCount] = proxyId
		} else {
			moveBuffer.add(proxyId)
		}

		moveCount++
	}

	private fun unBufferMove(proxyId: Int) {
		for (i in 0 until moveCount) {
			if (moveBuffer[i] == proxyId) {
				moveBuffer[i] = e_nullProxy
			}
		}
	}

	override fun query(aabb: AABB, callback: ProxyQueryCallback): Boolean {
		return tree.query(aabb, callback)
	}

	override fun rayCast(input: RayCastInput, callback: ProxyRayCastCallback) {
		tree.rayCast(input, callback)
	}

	override fun shiftOrigin(newOrigin: Vector2d) {
		tree.shiftOrigin(newOrigin)
	}

	override fun getUserData(proxyID: Int): Any? {
		return tree.getUserData(proxyID)
	}

	override fun getFatAABB(proxyID: Int): AABB {
		return tree.getFatAABB(proxyID)
	}

	/**
	 * Get the height of the embedded tree.
	 */
	val treeHeight: Int
		get() = tree.height

	/**
	 * Get the balance of the embedded tree.
	 */
	val treeBalance: Int
		get() = tree.maxBalance

	/**
	 * Get the quality metric of the embedded tree.
	 */
	val treeQuality: Double
		get() = tree.getAreaRatio

	/**
	 * Test overlap of fat AABBs.
	 */
	fun testOverlap(proxyIDA: Int, proxyIDB: Int): Boolean {
		return tree.getFatAABB(proxyIDA).intersect(tree.getFatAABB(proxyIDB))
	}

	private var queryProxy: Int = -1

	private fun queryCallback(proxyId: Int, userData: Any?): Boolean {
		if (proxyId == queryProxy) {
			return true
		}

		val moved = tree.wasMoved(proxyId)

		if (moved && proxyId > queryProxy) {
			// Both proxies are moving. Avoid duplicate pairs.
			return true
		}

		if (proxyId > queryProxy) {
			pairBuffer.add(queryProxy to proxyId)
		} else {
			pairBuffer.add(proxyId to queryProxy)
		}

		return true
	}

	/**
	 * Update the pairs. This results in pair callbacks. This can only add pairs.
	 */
	fun updatePairs(callback: (Any?, Any?) -> Unit) {
		pairBuffer.clear()

		for (i in 0 until moveCount) {
			val value = moveBuffer[i]

			if (value == e_nullProxy)
				continue

			val fatAABB = tree.getFatAABB(value)
			queryProxy = value
			tree.query(fatAABB, this::queryCallback)
		}

		// Send pairs to caller
		for (primaryPair in pairBuffer) {
			val userDataA = tree.getUserData(primaryPair.first)
			val userDataB = tree.getUserData(primaryPair.second)

			callback.invoke(userDataA, userDataB)
		}

		// Clear move flags
		for (i in 0 until moveCount) {
			val value = moveBuffer[i]

			if (value != e_nullProxy) {
				tree.clearMoved(value)
			}
		}

		// Reset move buffer
		moveCount = 0
	}
}
