package ru.dbotthepony.kbox2d.collision.shapes

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times

/**
 * A line segment (edge) shape. These can be connected in chains or loops
 * to other edge shapes. Edges created independently are two-sided and do
 * no provide smooth movement across junctions.
 */
class EdgeShape : IShape<EdgeShape> {
	override fun copy(): EdgeShape {
		return EdgeShape().also {
			it.oneSided = oneSided
			it.vertex0 = vertex0
			it.vertex1 = vertex1
			it.vertex2 = vertex2
			it.vertex3 = vertex3
		}
	}

	override val type: IShape.Type = IShape.Type.EDGE
	override val childCount: Int = 1

	override fun rayCast(input: RayCastInput, transform: Transform, childIndex: Int): RayCastOutput {
		// Put the ray into the edge's frame of reference.
		val p1 = b2MulT(transform.q, input.p1 - transform.p)
		val p2 = b2MulT(transform.q, input.p2 - transform.p)
		val d = p2 - p1

		val v1 = vertex1
		val v2 = vertex2
		val e = v2 - v1

		// Normal points to the right, looking from v1 at v2
		val normal = Vector2d(e.y, -e.x).normalized

		// q = p1 + t * d
		// dot(normal, q - v1) = 0
		// dot(normal, p1 - v1) + t * dot(normal, d) = 0
		val numerator = b2Dot(normal, v1 - p1)

		if (oneSided && numerator > 0.0) {
			return RayCastOutput.MISS
		}

		val denominator = b2Dot(normal, d)

		if (denominator == 0.0) {
			return RayCastOutput.MISS
		}

		val t = numerator / denominator

		if (t < 0.0 || input.maxFraction < t) {
			return RayCastOutput.MISS
		}

		val q = p1 + t * d

		// q = v1 + s * r
		// s = dot(q - v1, r) / dot(r, r)
		val r = v2 - v1
		val rr = b2Dot(r, r)

		if (rr == 0.0) {
			return RayCastOutput.MISS
		}

		val s = b2Dot(q - v1, r) / rr

		if (s < 0.0 || 1.0 < s) {
			return RayCastOutput.MISS
		}

		val outputNormal: Vector2d

		if (numerator > 0.0) {
			outputNormal = -b2Mul(transform.q, normal)
		} else {
			outputNormal = b2Mul(transform.q, normal)
		}

		return RayCastOutput(
			hit = true,
			normal = outputNormal,
			fraction = t
		)
	}

	override fun computeAABB(transform: Transform, childIndex: Int): AABB {
		val v1 = b2Mul(transform, vertex1)
		val v2 = b2Mul(transform, vertex2)

		val lower = b2Min(v1, v2)
		val upper = b2Max(v1, v2)

		val r = Vector2d(radius, radius)

		return AABB(
			mins = lower - r,
			maxs = upper + r
		)
	}

	override fun computeMass(density: Double): MassData {
		return MassData(
			center = 0.5 * (vertex1 + vertex2),
			mass = 0.0,
			inertia = 0.0
		)
	}

	override var radius: Double = b2_polygonRadius
		set(value) {
			if (value == b2_polygonRadius) {
				field = value
				return
			}

			throw UnsupportedOperationException("For polygonal shapes this must be b2_polygonRadius")
		}

	/**
	 * Set this as a part of a sequence. Vertex [v0] precedes the edge and vertex [v3]
	 * follows. These extra vertices are used to provide smooth movement
	 * across junctions. This also makes the collision one-sided. The edge
	 * normal points to the right looking from [v1] to [v2].
	 */
	fun setOneSided(v0: Vector2d, v1: Vector2d, v2: Vector2d, v3: Vector2d) {
		oneSided = true
		vertex0 = v0
		vertex1 = v1
		vertex2 = v2
		vertex3 = v3
	}

	/**
	 * Set this as an isolated edge. Collision is two-sided.
	 */
	fun setTwoSided(v1: Vector2d, v2: Vector2d) {
		oneSided = false
		vertex1 = v1
		vertex2 = v2
	}

	/// These are the edge vertices
	var vertex1 = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal vertex $value")
			}

			field = value
		}

	var vertex2 = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal vertex $value")
			}

			field = value
		}

	/// Optional adjacent vertices. These are used for smooth collision.
	var vertex0 = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal vertex $value")
			}

			field = value
		}

	var vertex3 = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal vertex $value")
			}

			field = value
		}

	var oneSided = false
}
