package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.dynamics.Body
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.util.EnumMap

fun interface JointFactory {
	fun factorize(jointDef: IJointDef): AbstractJoint
}

sealed class AbstractJoint(def: IJointDef) : IMovable {
	init {
		require(def.bodyA != def.bodyB) { "Tried to create join on same body" }
	}

	internal var isOnIsland = false

	protected var index: Int = 0

	/**
	 * Get collide connected.
	 * Note: modifying the collide connect flag won't work correctly because
	 * the flag is only checked when fixture AABBs begin to overlap.
	 */
	val collideConnected: Boolean = def.collideConnected

	/**
	 * Dump this joint to the log file.
	 */
	open fun dump() { }

	/**
	 * Debug draw this joint
	 */
	open fun draw(draw: IDebugDraw) { }

	/**
	 * Get the type of the concrete joint.
	 */
	val type: JointType = def.type
	var userData: Any? = def.userData

	// KBox2D: In original code, nothing expects bodies to be null
	// but certain joints (notably, mouse joint) have no meaningful
	// value for second body involved. So, KBox2D CAN handle case where
	// joint have only one body, and to avoid null assertions in all places
	// possible, bodyA and bodyB getters assert for null by themselves
	// However, there is nullable getter for bodies: nullableBodyA and nullableBodyB
	protected var _bodyA: Body? = def.bodyA as Body?
	protected var _bodyB: Body? = def.bodyB as Body?

	/**
	 * Get the first body attached to this joint.
	 */
	val bodyA: Body get() = checkNotNull(_bodyA) { "Body A is not present" }

	/**
	 * Get the second body attached to this joint.
	 */
	val bodyB: Body get() = checkNotNull(_bodyB) { "Body B is not present" }

	/**
	 * Get the anchor point on bodyA in world coordinates.
	 */
	abstract val anchorA: Vector2d

	/**
	 * Get the anchor point on bodyB in world coordinates.
	 */
	abstract val anchorB: Vector2d

	/**
	 * Get the reaction force on bodyB at the joint anchor in Newtons.
	 */
	abstract fun getReactionForce(inv_dt: Double): Vector2d

	/**
	 * Get the reaction torque on bodyB in N*m.
	 */
	abstract fun getReactionTorque(inv_dt: Double): Double

	val hasBodyA: Boolean get() = _bodyA != null
	val hasBodyB: Boolean get() = _bodyB != null
	val hasTwoBodies: Boolean get() = hasBodyA && hasBodyB

	val nullableBodyA get() = _bodyA
	val nullableBodyB get() = _bodyB

	private var _edgeA: JointEdge? = JointEdge(other = _bodyB, joint = this, next = _bodyA?.jointList)
	private var _edgeB: JointEdge? = JointEdge(other = _bodyA, joint = this, next = _bodyB?.jointList)

	internal val edgeA: JointEdge get() = checkNotNull(_edgeA) { "Edge A is not present" }
	internal val edgeB: JointEdge get() = checkNotNull(_edgeB) { "Edge B is not present" }

	var isValid: Boolean = true
		private set

	init {
		// Connect to the bodies' doubly linked lists.
		_bodyA?.jointList?.prev = edgeA
		_bodyA?.jointList = edgeA
		_bodyB?.jointList?.prev = edgeB
		_bodyB?.jointList = edgeB
	}

	/**
	 * Get the next joint the world joint list.
	 *
	 * Kotlin: This is not a list, but, as in C++ impl, a custom
	 * linked list.
	 */
	var next: AbstractJoint? = null
		internal set
	var prev: AbstractJoint? = null
		internal set

	fun destroy() {
		bodyA.world.destroyJoint(this)
	}

	/**
	 * Short-cut function to determine if either body is enabled.
	 */
	val isEnabled: Boolean get() = bodyA.isEnabled || bodyB.isEnabled

	/**
	 * Signals that this joint was destroyed, invalidate stuff
	 * to fail-fast this object
	 */
	internal open fun unlink() {
		_edgeA = null
		_edgeB = null
		_bodyA = null
		_bodyB = null
		isValid = false
	}

	internal abstract fun initVelocityConstraints(data: B2SolverData)
	internal abstract fun solveVelocityConstraints(data: B2SolverData)
	internal abstract fun solvePositionConstraints(data: B2SolverData): Boolean

	override fun shiftOrigin(newOrigin: Vector2d) {
		// no-op
	}

	companion object {
		private val registry = EnumMap<JointType, JointFactory>(JointType::class.java)

		internal fun register(jointType: JointType, factory: JointFactory) {
			require(registry.put(jointType, factory) == null) { "Re-registered $jointType factory" }
		}

		internal fun create(jointType: JointType, jointDef: IJointDef): AbstractJoint {
			return requireNotNull(registry[jointType]) { "No joint factory registered for type $jointType" }.factorize(jointDef)
		}

		internal fun create(jointDef: IJointDef): AbstractJoint {
			return create(jointDef.type, jointDef)
		}

		init {
			register(JointType.DISTANCE) {
				return@register DistanceJoint(it as DistanceJointDef)
			}

			register(JointType.REVOLUTE) {
				return@register RevoluteJoint(it as RevoluteJointDef)
			}

			register(JointType.PRISMATIC) {
				return@register PrismaticJoint(it as PrismaticJointDef)
			}

			register(JointType.PULLEY) {
				return@register PulleyJoint(it as PulleyJointDef)
			}

			register(JointType.GEAR) {
				return@register GearJoint(it as GearJointDef)
			}

			register(JointType.MOUSE) {
				return@register MouseJoint(it as MouseJointDef)
			}

			register(JointType.WHEEL) {
				return@register WheelJoint(it as WheelJointDef)
			}

			register(JointType.WELD) {
				return@register WeldJoint(it as WeldJointDef)
			}

			register(JointType.FRICTION) {
				return@register FrictionJoint(it as FrictionJointDef)
			}

			register(JointType.MOTOR) {
				return@register MotorJoint(it as MotorJointDef)
			}
		}
	}
}
