
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.api

import kotlin.math.sqrt

/**
 * Root interface of vector inheritance tree, providing all relevant methods
 * of operation on vectors of fixed length
 */
interface IVector<T : IVector<T>> {
	/**
	 * Length of this vector, as Double
	 */
	val length: Double get() = sqrt(lengthSquared)

	/**
	 * Squared length of this vector, as Double
	 */
	val lengthSquared: Double

	/**
	 * Whenever all components of vector are finite
	 *
	 * Keep in mind that in some situations infinite vectors are still making sense
	 * (e.g. creating [AABB] that cover entire game, or alike, world)
	 */
	val isFinite: Boolean

	/**
	 * Whenever at least one of components of vector are NaN
	 */
	val isNaN: Boolean

	/**
	 * Whenever is this vector valid (components are not NaN and are finite)
	 *
	 * Keep in mind that in some situations infinite vectors are still making sense
	 * (e.g. creating [AABB] that cover entire game, or alike, world)
	 */
	val isValid: Boolean get() = isFinite && !isNaN

	/**
	 * Adds components of both vectors, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun plus(other: T): T

	/**
	 * Subtracts components of both vectors, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun minus(other: T): T

	/**
	 * Multiplies components of both vectors, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun times(other: T): T

	/**
	 * Divides components of both vectors, returning new vector as result
	 *
	 * @throws ArithmeticException if vector work with whole numbers, and
	 * one of [other] components end up zero
	 *
	 * @return new vector
	 */
	operator fun div(other: T): T

	/**
	 * New vector, with every component being absolute
	 */
	val absoluteValue: T

	/**
	 * Takes all components of both vectors, and choose the smallest of each,
	 * and construct a new vector from them
	 *
	 * @return new vector
	 */
	fun coerceAtMost(maximal: T): T

	/**
	 * Takes all components of both vectors, and choose the biggest of each,
	 * and construct a new vector from them
	 *
	 * @return new vector
	 */
	fun coerceAtLeast(minimal: T): T

	/**
	 * Takes all components of this vector, and clamp them in between [minimal] and [maximal] corresponding
	 * components and construct a new vector from them
	 *
	 * @return new vector
	 */
	fun clamp(minimal: T, maximal: T): T

	/**
	 * Constructs a new vector with all components negated
	 *
	 * @return new vector
	 */
	operator fun unaryMinus(): T

	/**
	 * Returns square rooted distance bewteen this and [other]
	 *
	 * @return distance between vectors (as points) as [Double]
	 */
	fun distance(other: T): Double {
		return sqrt(distanceSquared(other))
	}

	/**
	 * Returns squared distance between this and [other]
	 *
	 * @return distance between vectors (as points) as [Double]
	 */
	fun distanceSquared(other: T): Double
}

/**
 * Define all relatable methods on vectors which components are fractional
 */
interface IFractionalVector<T : IFractionalVector<T>> {
	/**
	 * New (directional) vector, with length of 1
	 *
	 * If vector length is zero, a zero vector is returned
	 */
	val normalized: T
}

/**
 * Define all relatable mutating methods on vectors which components are fractional
 */
interface IMutableFractionalVector<T : IMutableFractionalVector<T>> {
	/**
	 * Normalizes this vector (by setting all of its components) to have 1 unit length
	 * (turning it into directional or unit vector),
	 * returning old length as [Double]
	 *
	 * @return old length as [Double]
	 */
	fun normalize(): Double
}

/**
 * Define all relatable methods on vectors which components are whole numbers
 */
interface IWholeVector<T : IWholeVector<T>> {
	/**
	 * Returns squared distance between this and [other], but as whole number
	 *
	 * @return distance between vectors (as points) as [Long]
	 */
	fun wholeDistanceSquared(other: T): Long
}

/**
 * Define all relatable mutating methods on vectors which components are whole numbers
 */
interface IMutableWholeVector<T : IMutableWholeVector<T>> {

}

/**
 * Define mutable vector x vector operations
 */
interface IMutableVector<T : IMutableVector<T, V>, V> {
	/**
	 * Adds components of both vectors, setting result to this vector
	 *
	 * @return this vector
	 */
	fun plusMut(other: V): T

	/**
	 * Subtracts components of both vectors, setting result to this vector
	 *
	 * @return this vector
	 */
	fun minusMut(other: V): T

	/**
	 * Multiplies components of both vectors, setting result to this vector
	 *
	 * @return this vector
	 */
	fun timesMut(other: V): T

	/**
	 * Divides components of both vectors, setting result to this vector
	 *
	 * @throws ArithmeticException if vector work with whole numbers, and
	 * one of [other] components end up zero
	 *
	 * @return this vector
	 */
	fun divMut(other: V): T
}

/**
 * Define mutable scalar x vector operations using [Double]s
 */
interface IMutableDoubleVector<T : IMutableDoubleVector<T>> {
	/**
	 * Multiplies components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun timesMut(other: Double): T

	/**
	 * Divides components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun divMut(other: Double): T
}

/**
 * Define scalar x vector operations using [Double]s
 */
interface IDoubleVector<T : IDoubleVector<T>> {
	/**
	 * Multiplies [other] by all components, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun times(other: Double): T

	/**
	 * Divides components by [other], returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun div(other: Double): T
}

/**
 * Define mutable scalar x vector operations using [Float]s
 */
interface IMutableFloatVector<T : IMutableFloatVector<T>> {
	/**
	 * Multiplies components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun timesMut(other: Float): T

	/**
	 * Divides components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun divMut(other: Float): T
}

/**
 * Define scalar x vector operations using [Float]s
 */
interface IFloatVector<T : IFloatVector<T>> {
	/**
	 * Multiplies [other] by all components, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun times(other: Float): T

	/**
	 * Divides components by [other], returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun div(other: Float): T
}

/**
 * Define mutable scalar x vector operations using [Int]s
 */
interface IMutableIntVector<T : IMutableIntVector<T>> {
	/**
	 * Multiplies components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun timesMut(other: Int): T

	/**
	 * Divides components of vector by [other], setting result to this vector
	 *
	 * @return this vector
	 */
	fun divMut(other: Int): T
}

/**
 * Define scalar x vector operations using [Int]s
 */
interface IIntVector<T : IIntVector<T>> {
	/**
	 * Multiplies [other] by all components, returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun times(other: Int): T

	/**
	 * Divides components by [other], returning new vector as result
	 *
	 * @return new vector
	 */
	operator fun div(other: Int): T
}

/**
 * Root class of vector inheritance tree, providing some concrete methods
 * for vectors of fixed length
 */
abstract class AbstractVector<T : AbstractVector<T>> : IMatrixLike, IVector<T> {
	final override val columns: Int = 1

	/**
	 * @throws IllegalArgumentException if vector is not finite
	 */
	inline fun requireIsFinite(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalArgumentException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalArgumentException if vector is not finite
	 */
	inline fun requireIsFinite() {
		if (!isFinite) {
			throw IllegalArgumentException("Vector is not finite")
		}
	}

	/**
	 * @throws IllegalStateException if vector is not finite
	 */
	inline fun checkIsFinite(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalStateException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if vector is not finite
	 */
	inline fun checkIsFinite() {
		if (!isFinite) {
			throw IllegalStateException("Vector is not finite")
		}
	}

	/**
	 * @throws IllegalArgumentException if vector has NaN component
	 */
	inline fun requireNotNaN(lambda: () -> Any) {
		if (isNaN) {
			throw IllegalArgumentException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalArgumentException if vector has NaN component
	 */
	inline fun requireNotNaN() {
		if (isNaN) {
			throw IllegalArgumentException("Vector has NaN component")
		}
	}

	/**
	 * @throws IllegalStateException if vector has NaN component
	 */
	inline fun checkNotNaN(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalStateException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if vector has NaN component
	 */
	inline fun checkNotNaN() {
		if (!isFinite) {
			throw IllegalStateException("Vector has NaN component")
		}
	}

	/**
	 * @throws IllegalArgumentException if vector has NaN or infinite component
	 */
	inline fun requireIsValid() {
		requireIsFinite()
		requireNotNaN()
	}

	/**
	 * @throws IllegalStateException if vector has NaN or infinite component
	 */
	inline fun checkIsValid() {
		requireIsFinite()
		requireNotNaN()
	}

	/**
	 * @throws IllegalArgumentException if vector has NaN or infinite component
	 */
	inline fun requireIsValid(lambda: () -> Any) {
		requireIsFinite(lambda)
		requireNotNaN(lambda)
	}

	/**
	 * @throws IllegalStateException if vector has NaN or infinite component
	 */
	inline fun checkIsValid(lambda: () -> Any) {
		requireIsFinite(lambda)
		requireNotNaN(lambda)
	}
}
