package ru.dbotthepony.kvector.matrix

import ru.dbotthepony.kvector.api.IMatrixGetterFloat
import ru.dbotthepony.kvector.matrix.nfloat.MutableMatrix4f
import ru.dbotthepony.kvector.vector.nfloat.Vector3f
import ru.dbotthepony.kvector.vector.nfloat.Vector4f

/**
 * This class represent Matrix stack using [MutableMatrix4f]. It is useful for 3D
 * rendering because it allows to create single stack, push new matrix in any
 * applicable context, pass stack to all rendering methods which will further push
 * transformation matrices, etc.
 *
 * All operations are mutating last matrix in stack.
 */
class Matrix4fStack {
	private val stack = ArrayDeque<MutableMatrix4f>()

	/**
	 * Last matrix in stack, you are free to modify it,
	 * but if you need to do contained modifications, such
	 * as inside your render subroutine (e.g. inside your entity drawer),
	 * please do use [push] first with no arguments, perform (render and) matrix operations,
	 * and then [pop] it.
	 *
	 * @return [MutableMatrix4f]
	 */
	val last get() = stack.last()

	init {
		stack.add(MutableMatrix4f())
	}

	/**
	 * Pops last matrix from stack.
	 *
	 * @return this stack
	 */
	fun pop(): Matrix4fStack {
		stack.removeLast()
		return this
	}

	/**
	 * Pushes matrix to stack.
	 * By default, pushes a copy of last matrix, which is desirable in most cases.
	 *
	 * @return this stack
	 */
	fun push(matrix: MutableMatrix4f = last.toMutableMatrix4f()): Matrix4fStack {
		stack.add(matrix)
		return this
	}

	/**
	 * Clears stack and pushes specified [matrix4f] into it as first element.
	 * By default, pushes identity matrix.
	 *
	 * @return this stack
	 */
	fun clear(matrix4f: MutableMatrix4f = MutableMatrix4f()): Matrix4fStack {
		stack.clear()
		stack.add(matrix4f)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.plusMut] on last matrix.
	 *
	 * @return this stack
	 */
	operator fun plus(other: IMatrixGetterFloat): Matrix4fStack {
		last.plusMut(other)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.minusMut] on last matrix.
	 *
	 * @return this stack
	 */
	operator fun minus(other: IMatrixGetterFloat): Matrix4fStack {
		last.minusMut(other)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.timesMut] on last matrix.
	 *
	 * @return this stack
	 */
	operator fun times(other: IMatrixGetterFloat): Matrix4fStack {
		last.timesMut(other)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.scaleMut] on last matrix.
	 *
	 * @return this stack
	 */
	fun scale(vector: Vector4f): Matrix4fStack {
		last.scaleMut(vector)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.scaleMut] on last matrix.
	 * This is a convenience function, as it just create new [Vector4f].
	 *
	 * @return this stack
	 */
	fun scale(x: Float = 1f, y: Float = 1f, z: Float = 1f, w: Float = 1f): Matrix4fStack {
		last.scaleMut(Vector4f(x, y, z, w))
		return this
	}

	/**
	 * Calls [MutableMatrix4f.translate] on last matrix.
	 *
	 * @return this stack
	 */
	fun translate(vector: Vector3f): Matrix4fStack {
		last.translateMut(vector)
		return this
	}

	/**
	 * Sets [MutableMatrix4f.translation] on last matrix.
	 * This is a convenience function, as it just create new [Vector3f].
	 *
	 * @return this stack
	 */
	fun translate(x: Float = 0f, y: Float = 0f, z: Float = 0f): Matrix4fStack {
		last.translateMut(Vector3f(x, y, z))
		return this
	}

	/**
	 * Calls [MutableMatrix4f.translateWithMultiplicationMut] on last matrix.
	 *
	 * @return this stack
	 */
	fun translateWithMultiplication(vector: Vector3f): Matrix4fStack {
		last.translateWithMultiplicationMut(vector)
		return this
	}

	/**
	 * Calls [MutableMatrix4f.translateWithMultiplicationMut] on last matrix.
	 * This is a convenience function, as it just create new [Vector3f].
	 *
	 * @return this stack
	 */
	fun translateWithMultiplication(x: Float = 0f, y: Float = 0f, z: Float = 0f): Matrix4fStack {
		last.translateWithMultiplicationMut(Vector3f(x, y, z))
		return this
	}

	/**
	 * Removes last matrix from stack, and pushes [matrix] to stack.
	 *
	 * @return this stack
	 */
	fun replace(matrix: MutableMatrix4f): Matrix4fStack {
		stack.removeLast()
		stack.add(matrix)
		return this
	}
}
