
@file:Suppress("unchecked_cast", "unused")

package ru.dbotthepony.kvector.matrix.nfloat

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.matrix.*
import ru.dbotthepony.kvector.matrix.generated.MultiplicationsFloat
import ru.dbotthepony.kvector.narray.Float2Dimensional

private fun checkRows(value: Int): Int {
	require(value > 0) { "Invalid amount of rows $value" }
	return value
}

private fun checkColumns(value: Int): Int {
	require(value > 0) { "Invalid amount of columns $value" }
	return value
}

/**
 * Abstract class with only one protected abstract method: [factorize].
 *
 * This class is meant to be only internally overridden (to implement interfaces), but if you really need to, you can
 * extend this class.
 */
abstract class AbstractMatrixVf<T : AbstractMatrixVf<T>> protected constructor(
	final override val columns: Int,
	final override val rows: Int,
	@JvmField protected val memory: Float2Dimensional
) : IMatrix<T>, IFloatMatrix<T>, ITransposable<T> {
	/**
	 * To avoid middle-object creation when deriving from [AbstractMatrixVf] ([MatrixVf]), this method
	 * allows to factory produce new instances of final class
	 */
	protected abstract fun factorize(memory: Float2Dimensional): T
	protected open val flexible = true

	final override fun get(column: Int, row: Int): Float {
		return memory[column, row]
	}

	override fun plus(other: IMatrixGetterFloat): T {
		requireSizeEquals(other)

		val result = Float2Dimensional(columns, rows)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				result[column, row] = memory[column, row] + other[column, row]
			}
		}

		return factorize(result)
	}

	override fun minus(other: IMatrixGetterFloat): T {
		requireSizeEquals(other)

		val result = Float2Dimensional(columns, rows)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				result[column, row] = memory[column, row] - other[column, row]
			}
		}

		return factorize(result)
	}

	override fun times(other: IMatrixGetterFloat): T {
		require(flexible || sizeEquals(other)) { "${this::class.qualifiedName} is not a flexible" }
		val result = multiplyMatrix(this, other)
		return factorize(result)
	}

	override fun times(other: Float): T {
		val result = Float2Dimensional(columns, rows)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				result[column, row] = memory[column, row] * other
			}
		}

		return factorize(result)
	}

	override fun div(other: Float): T {
		val result = Float2Dimensional(columns, rows)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				result[column, row] = memory[column, row] / other
			}
		}

		return factorize(result)
	}

	override val transposed: T
		get() {
			val result = transposeMatrix(this)
			return factorize(result)
		}

	override val trace: Float?
		get() {
			if (!isSquareMatrix)
				return null

			var calc = 0f

			for (i in 0 until rows) {
				calc += memory[columns, rows]
			}

			return calc
		}

	override val determinant: Float?
		get() {
			if (!isSquareMatrix)
				return null

			return matrixDeterminant(this)
		}

	override val isFinite: Boolean
		get() {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					if (!memory[column, row].isFinite()) {
						return false
					}
				}
			}

			return true
		}

	override val isNaN: Boolean
		get() {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					if (memory[column, row].isNaN()) {
						return true
					}
				}
			}

			return false
		}

	override val inverse: T?
		get() {
			if (!isSquareMatrix)
				return null

			val result = inverseMatrix(memory) ?: return null
			return factorize(result)
		}

	override val cofactor: T?
		get() {
			if (!isSquareMatrix)
				return null

			return factorize(cofactorMatrix(memory))
		}

	override val adjugate: T?
		get() {
			if (!isSquareMatrix)
				return null

			return factorize(adjugateMatrix(memory))
		}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false
		other as AbstractMatrixVf<*>

		if (other.columns != columns || other.rows != rows) {
			return false
		}

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				if (memory[column, row] != other.memory[column, row]) {
					return false
				}
			}
		}

		return true
	}

	override fun hashCode(): Int {
		if (columns == 0 || rows == 0) {
			return 0
		}

		var result = 0

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				result = 31 * result + memory[column, row].hashCode()
			}
		}

		return result
	}

	override fun toString(): String {
		val builder = StringBuilder("[\n")

		val strings = arrayOfNulls<String>(columns * rows)

		var width = 0
		val columnWidths = IntArray(columns)

		for (column in 0 until columns) {
			var columnWidth = 0

			for (row in 0 until rows) {
				strings[column + row * columns] = memory[row, column].toString()
				columnWidth = columnWidth.coerceAtLeast(strings[column + row * columns]!!.length)
			}

			columnWidths[column] = columnWidth + 1
			width += columnWidth + 1
		}

		for (row in 0 until rows) {
			builder.append("    ")

			for (column in 0 until columns) {
				builder.append(strings[column + row * columns]!!)

				if (column == columns - 1) {
					builder.append(";")
				} else {
					builder.append(",")
				}

				builder.append(" ".repeat(columnWidths[column] - strings[column + row * columns]!!.length))
			}

			builder.append("\n")
		}

		builder.append("]")

		return builder.toString()
	}
}

/**
 * Abstract class inheriting [AbstractMatrixVf], implementing mutating methods.
 *
 * This class is meant to be only internally overridden (to implement interfaces), but if you really need to, you can
 * extend this class.
 */
abstract class AbstractMutableMatrixVf<T : AbstractMutableMatrixVf<T>> protected constructor(
	columns: Int,
	rows: Int,
	memory: Float2Dimensional
) : AbstractMatrixVf<T>(columns, rows, memory), IMutableFloatMatrix<T>, IMutableTransposable<T> {
	final override fun set(column: Int, row: Int, value: Float) {
		memory[column, row] = value
	}

	override fun timesMut(other: Float): T {
		for (column in 0 until columns) {
			for (row in 0 until rows) {
				memory[column, row] *= other
			}
		}

		return this as T
	}

	override fun divMut(other: Float): T {
		for (column in 0 until columns) {
			for (row in 0 until rows) {
				memory[column, row] /= other
			}
		}

		return this as T
	}

	override fun timesMut(other: IMatrixGetterFloat): T {
		requireSizeEquals(other)

		if (MultiplicationsFloat.multiplyMatrix(this, other, this) == null) {
			val result = multiplyMatrix(this, other)

			for (column in 0 until columns) {
				for (row in 0 until rows) {
					memory[column, row] = result[column, row]
				}
			}
		}

		return this as T
	}

	override fun plusMut(other: IMatrixGetterFloat): T {
		requireSizeEquals(other)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				memory[column, row] += other[column, row]
			}
		}

		return this as T
	}

	override fun minusMut(other: IMatrixGetterFloat): T {
		requireSizeEquals(other)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				memory[column, row] -= other[column, row]
			}
		}

		return this as T
	}

	override fun transpose(): T {
		check(isSquareMatrix) { "This matrix is not a square matrix (${columns}x${rows})" }

		val result = transposeMatrix(this)

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				memory[column, row] = result[column, row]
			}
		}

		return this as T
	}
}

/**
 * Represents immutable matrix with arbitrary dimensions, storing values as [Float]s,
 * backed by [Float2Dimensional].
 *
 * Matrix is initialized to zeros upon creation.
 */
open class MatrixVf protected constructor(
	columns: Int,
	rows: Int,
	memory: Float2Dimensional
) : AbstractMatrixVf<MatrixVf>(columns, rows, memory) {
	constructor(columns: Int, rows: Int) : this(checkColumns(columns), checkRows(rows), Float2Dimensional(checkColumns(columns), checkRows(rows)))

	override fun factorize(memory: Float2Dimensional): MatrixVf {
		return MatrixVf(memory.columns, memory.rows, memory)
	}

	companion object {
		/**
		 * Copies specified matrix into new [MatrixVf]
		 */
		fun copy(input: IMatrixGetterFloat): MatrixVf {
			val memory = Float2Dimensional(input.columns, input.rows)

			for (column in 0 until input.columns) {
				for (row in 0 until input.rows) {
					memory[column, row] = input[column, row]
				}
			}

			return MatrixVf(input.columns, input.rows, memory)
		}
	}
}

/**
 * Represents mutable matrix with arbitrary dimensions, storing values as [Float]s,
 * backed by [Float2Dimensional].
 *
 * Matrix is initialized to zeros upon creation.
 */
open class MutableMatrixVf protected constructor(
	columns: Int,
	rows: Int,
	memory: Float2Dimensional
) : AbstractMutableMatrixVf<MutableMatrixVf>(columns, rows, memory) {
	constructor(columns: Int, rows: Int) : this(checkColumns(columns), checkRows(rows), Float2Dimensional(checkColumns(columns), checkRows(rows)))

	override fun factorize(memory: Float2Dimensional): MutableMatrixVf {
		return MutableMatrixVf(memory.columns, memory.rows, memory)
	}

	companion object {
		/**
		 * Copies specified matrix into new [MutableMatrixVf]
		 */
		fun copy(input: IMatrixGetterFloat): MutableMatrixVf {
			val memory = Float2Dimensional(input.columns, input.rows)

			for (column in 0 until input.columns) {
				for (row in 0 until input.rows) {
					memory[column, row] = input[column, row]
				}
			}

			return MutableMatrixVf(input.columns, input.rows, memory)
		}
	}
}
